#include <mpi.h>
#include <omp.h>

#include "heterogeneous_FDM_config.h"

real_t cpu_kernel_without_intra_border ( real_t ** __restrict__ h_u, int_t gpucpuh, int_t gpuh, int_t width){
    struct timeval cpuA, cpuB;
    gettimeofday(&cpuA, NULL);
    #pragma omp parallel for
    for ( int_t y=gpuh+1; y<gpucpuh; y++ ){
        for ( int_t x=0; x<width; x++ ){   
            // load stencil values
            real_t east, south, west, north, center0, center1;
            center0 = H_U(0,y,x);
            center1 = H_U(1,y,x);
            west =  (x == 0             ? H_U(1, y,x+1) : H_U(1, y,x-1));
            east =  (x == sidelength-1  ? H_U(1, y,x-1) : H_U(1, y,x+1));
            north = (y == 0             ? H_U(1, y+1,x) : H_U(1, y-1,x));
            south = (y == sidelength-1  ? H_U(1, y-1,x) : H_U(1, y+1,x));

            H_U(2,y,x) = 2.0*center1 - center0 + (c*c) * (-4.0f*center1 + east + south + north + west);
        }
    }
    gettimeofday(&cpuB, NULL);
    return time_diff_in_ms(cpuA, cpuB);
}

real_t cpu_inner_kernel ( real_t ** __restrict__ h_u ){
    struct timeval cpuA, cpuB;
    gettimeofday(&cpuA, NULL);
    #pragma omp parallel for
    for ( int_t y=local_size[GPU][Y]+1; y<local_size[GPUANDCPU][Y]-1; y++ ){
        for ( int_t x=0; x<local_size[CPU][X]; x++){   
            // load stencil values
            real_t east, south, west, north, center0, center1;
            center0 = H_U(0,y,x);
            center1 = H_U(1,y,x);
            west =  (x == 0                     ? H_U(1, y,x+1) : H_U(1, y,x-1));
            east =  (x == local_size[CPU][X]-1  ? H_U(1, y,x-1) : H_U(1, y,x+1));
            north = H_U(1, y-1,x);
            south = H_U(1, y+1,x);

            H_U(2,y,x) = 2.0*center1 - center0 + (c*c) * (-4.0f*center1 + east + south + north + west);
        }
    }
    gettimeofday(&cpuB, NULL);
    return time_diff_in_ms(cpuA, cpuB);
}

void cpu_intra_border_kernel ( void ){
    int_t y=local_size[GPU][Y];
    int_t gpucpuheight = local_size[GPUANDCPU][Y];
    #pragma omp parallel for
    for ( int_t x=0; x<sidelength; x++ ){   
        // load stencil values
        real_t east, south, west, north, center0, center1;
        center0 = H_U(0,y,x);
        center1 = H_U(1,y,x);
        west =  (x == 0             ? H_U(1, y,x+1) : H_U(1, y,x-1));
        east =  (x == sidelength-1  ? H_U(1, y,x-1) : H_U(1, y,x+1));
        north = (y == 0             ? H_U(1, y+1,x) : H_U(1, y-1,x));
        south = (y == gpucpuheight-1  ? H_U(1, y-1,x) : H_U(1, y+1,x));

        H_U(2,y,x) = 2.0*center1 - center0 + (c*c) * (-4.0f*center1 + east + south + north + west);
    }
}

void cpu_border_kernel ( void ){
    int_t width = local_size[CPU][X];
    int_t gpuheight = local_size[GPU][Y];
    int_t subgridheight = local_size[GPUANDCPU][Y];
    #pragma omp parallel for
    for ( int_t i=0; i<width*2; i++ ){   
        int_t x=0, y=0;
        if (i < width){   // northern border
            y = gpuheight;
            x = i;
        }
        else if (i < 2*width){    // southern border
            y = subgridheight - 1;
            x = i-width;
        }
        real_t mid1   = H_U(1, y,x);
        real_t east1  = (x==width-1  ? H_U(1, y, x-1) : H_U(1, y, x+1));
        real_t west1  = (x==0        ? H_U(1, y, x+1) : H_U(1, y, x-1));
        real_t north1 = (north< 0 && y==0        ? H_U(1, y+1, x) : H_U(1, y-1, x));
        real_t south1 = (south< 0 && y==subgridheight-1 ? H_U(1, y-1, x) : H_U(1, y+1, x));

        H_U(2, y,x) = 2.0f*mid1 - H_U(0, y,x) + (c*c) * (-4.0f*mid1 + east1 + south1 + north1 + west1);
    }
}