#include "heterogeneous_FDM_config.h"

// set the size of the grid, shoud be set -------
#ifdef SIZE
    const int sidelength = SIZE;
#else
    const int sidelength = 128;
    #warning "DSIZE flag missing, using default value..."
#endif

// decide whether or not to animate -------------
#ifdef ANIMATE
    const bool animate = true;
#else
    const bool animate = false;
#endif

// set animation parameters ---------------------
#ifdef SNAP_FREQ
    const int snap_freq = SNAP_FREQ;
#else
    const int snap_freq = 200;
    #warning "DSNAP_FREQ flag missing, using default value..."
#endif

#ifdef ITERATIONS
    const int_t max_iter = ITERATIONS;
#else
    const int_t max_iter = 40000;
    #warning "DITERATIONS flag missing, using default value..."
#endif

#ifdef INTER_CALIBRATE
    const bool inter_calibrate=INTER_CALIBRATE;
#else
    const bool inter_calibrate=false;
#endif

// default value is 1.0f if it is not set, but it can be overwritten to 0.5 if the user wants to calibrate it
// 0.5 is chosen as a middle point to start a binary search
#ifndef GPUSHARE
#define GPUSHARE 1.0f
#endif

#ifdef BSEARCH_CALIBRATION
    const bool intra_calibrate = true;
    #undef GPUSHARE
    #define GPUSHARE 0.50f;
#else
    const bool intra_calibrate = false;
#endif

real_t gpu_work_share = GPUSHARE;

real_t time_diff_in_ms(struct timeval t1, struct timeval t2) {
    real_t diff_sec = t2.tv_sec - t1.tv_sec;
    real_t diff_usec = t2.tv_usec - t1.tv_usec;
    return (diff_sec * 1000.0f) + (diff_usec / 1000.0f);
}

// custom structs
void binary_search_handler_setup(struct binary_search_handler* bin, int_t l, int_t r, bool done){
    bin->l = l;
    bin->r = r;
    bin->mid = (l+r)/2;
    bin->bisections_made = 0;
    bin->converged = done;
}

void binary_search(struct binary_search_handler* bin, real_t cputime, real_t gputime){
    if (cputime != -1. && gputime != -1. && bin->l <= bin->r){
        if (cputime > gputime){ // slow CPU -> increase amount of GPU area
            bin->l = bin->mid + 1;
            bin->mid = (bin->l + bin->r)/2;
        }
        else{ // slow GPU -> incrase amount of CPU area
            bin->r = bin->mid - 1;
            bin->mid = (bin->l + bin->r)/2;
        }

        if (bin->l > bin->r){
            gpu_work_share = (real_t)bin->mid / (real_t) local_size[GPUANDCPU][Y];
            printf("converged gpu_work_share: %f\n", gpu_work_share);
            bin->converged = true;
        }
    }
}

const float dt = 1.0;
const float h  = 1e-2;  
const float c = (1e-2)/1.0;

// [GPUDATA, CPUDATA, GPU_AND_CPU]
int
    rank,
    size,
    north, south,
    global_size[2],
    local_size[3][2],
    local_origin[3][2],
    mpi_circumference,
    local_size_bytes[3];
char ns;

real_t
    *h_u[3],
    *d_u0,
    *d_u1,
    *d_u2,
    *send_borders,
    *recv_borders,
    row_per_s=1.0; 