#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>

// create types for integers and real numbers ---
typedef int64_t int_t;
typedef float real_t;

// set simulation parameters --------------------
extern const real_t dt, h, c;

// indexing shortcuts for more readability
#define Y 0
#define X 1
enum device{GPU, CPU, GPUANDCPU};
enum kernel_type{FULL, INNER, FULL_WITHOUT_INTRA_BORDER};
enum border_kernel_type{INTRA_BORDER, DEVICE_BORDER, INTER_BORDER};
enum sync_type{DEVICE, BORDER_DONE, PACKING_DONE};

// custom structs
struct binary_search_handler{
    int_t l, r, mid, bisections_made, frequency;
    bool converged;
};
extern void binary_search_handler_setup(struct binary_search_handler*, int_t, int_t, bool);
extern void binary_search(struct binary_search_handler*, real_t, real_t);

// shortcute defines
#define MIN(a,b) ((a)<=(b) ? (a) : (b))
#define MAX(a,b) ((a)>=(b) ? (a) : (b))
#define CEIL_DIV(a,d) (((a)+(d)-1)/(d)) // modular arithmetic trick to ceiling divide 

#define H_U(t,y,x) h_u[(t)][((y)+1)*(local_size[GPUANDCPU][X]+2)+((x)+1)]

#define D_U0(y,x) d_u0[((y)+1)*(2+width)+(x)+1]
#define D_U1(y,x) d_u1[((y)+1)*(2+width)+(x)+1]
#define D_U2(y,x) d_u2[((y)+1)*(2+width)+(x)+1]

// bitmask for seeing whether an mpi process in on the edge of the plane
// a bit in nesw should be true if we are on the edge using the following order nesw[north, east, south, west]
#define ON_N_EDGE (ns&1)
#define ON_S_EDGE (ns&2)

// global variables that should only be defines in a single file
extern const int sidelength;
extern const bool animate;
extern const int snap_freq;
extern const bool inter_calibrate;
extern const bool intra_calibrate;
extern const int_t max_iter;
extern real_t gpu_work_share;

extern int
    rank,
    size,
    north, south,
    global_size[2],
    local_size[3][2],
    local_origin[3][2],
    mpi_circumference,
    local_size_bytes[3];
extern char ns;

extern real_t
    *h_u[3],
    *d_u0,
    *d_u1,
    *d_u2,
    *send_borders,
    *recv_borders,
    row_per_s;

// main functions
void update_global_variables_with_new_gpu_share(int_t);
void update_global_variables_with_new_subdomain_size(real_t);
void set_mutable_global_variables_int ( int_t );
void set_mutable_global_variables_real ( real_t );
void set_immutable_global_variables();
void initialize_global_mem( void );
void initialize_MPI_types ( void );
void inter_border_exchange ( bool );
void save_state ( int_t iter );
void init_mpi( void );
void rotate_buffers();
void distribute_local_shares();

// cuda functions 
void init_cuda();
void update_local_dimensions();
void inter_calibration(real_t*);
void inter_and_intra_calibration(real_t*);
void get_gputime(real_t*);
void cuda_sync(enum sync_type);
void initialize_cuda_mem();
void initialize_state_with_gpu();
void set_cuda_grids();
void cuda_terminate();
void fetch_gpu_grids ( int_t );
void intra_border_exchange();
void gpu_pack_borders(enum border_kernel_type);
void gpu_unpack_borders(enum border_kernel_type);
void gpu_border_kernel(enum border_kernel_type);
void gpu_main_kernel(enum kernel_type);

// cpu functions
real_t cpu_kernel_without_intra_border(real_t ** __restrict__, int_t, int_t, int_t);
void cpu_intra_border_kernel();
void cpu_border_kernel();
real_t cpu_inner_kernel(real_t** __restrict__);

// helper functions
real_t time_diff_in_ms(struct timeval, struct timeval);